import click

from blazetest.cli.base import cli
from blazetest.core.project_config.project_config import ProjectConfiguration
from blazetest.core.run_test.result_model import TestSessionResultManager
from blazetest.core.utils.s3_manager import S3Manager
from blazetest.core.utils.utils import print_table

SESSIONS_TABLE_COLUMNS = [
    "UUID",
    "RERUN UUID",
    "STACK",
    "EXECUTED AT",
    "PASSED TESTS",
    "TAGS",
]


@cli.command()
@click.option(
    "-t",
    "--tags",
    help="Option to filter the sessions by tags. If not specified, all results are shown.",
)
@click.option(
    "-i",
    "--include",
    default="all",
    type=click.Choice(["all", "runs", "reruns"]),
    help="Which type of runs to include in the output of the history of sessions. Can be: all, runs, reruns. "
    "Default is all",
)
@click.option(
    "-config",
    "--config-path",
    help="Configuration path to the TOML file for the CLI. "
    "If not specified -> searches the project's root folder for the file blazetest.toml. "
    "If not found -> raises an error.",
)
def sessions(tags: str = None, include: str = "all", config_path: str = None):
    """
    Shows the sessions created by Blazetest CLI.
    """
    config = ProjectConfiguration.from_toml_file(config_path)
    table = [SESSIONS_TABLE_COLUMNS]

    s3_manager = S3Manager(aws_config=config.cloud.aws)
    _ = s3_manager.find_s3_bucket()

    manager = TestSessionResultManager(
        config=config,
        s3_manager=s3_manager,
    )

    results = manager.get_test_session_results(tags=tags, include=include)

    for result in results:
        table.append(result.get_tabular_data())

    print_table(table)



import base64
import json
from typing import Dict

import pulumi
import pulumi_aws as aws
from pulumi_docker import RegistryArgs, DockerBuildArgs, Image


class AWSWorkflow:
    """
    The Workflow class is used to deploy an image to Amazon Elastic Container Registry (ECR)
    and an AWS Lambda function to an AWS CloudFormation stack.

    Attributes:
        project_path (str): The path to the project directory which will be the context for Docker image.
        docker_file_path (str): The path to the Dockerfile which will be used to build the project.
        ecr_repository_prefix (str): The name of the ECR repository to deploy the image to.
        s3_bucket_name (str): The name of the S3 bucket to use.
        resource_prefix (str): The name of the CloudFormation stack to deploy the Lambda function to.
        loki_user (str): Loki username
        loki_host (str): Loki host address name
        loki_api_key (str): Loki API Key
        tags (str): Tags to pass to the lambda

    Properties:
        ecr_repository (aws.ecr.Repository): The ECR repository object.
        lambda_function (aws.lambda_.Function): The Lambda function object.
        image (Image): The image object.
        s3_bucket_prefix (aws.s3.Bucket): The S3 bucket object.
        env_vars (Dict): A dictionary of environment variables to set for the Lambda function.

    Methods:
        deploy(): Initializes the ECR repository, S3 bucket, image, and Lambda function, and deploys them.

        create_s3_bucket(): Creates the S3 bucket.
        create_ecr_repository(): Creates the ECR repository.
        create_ecr_repository_image(): Creates the image and builds it using Docker.
        create_lambda_function(): Creates the Lambda function.

        get_lambda_iam_policy(): Retrieves IAM Policy for attaching to IAM Role
        get_lambda_iam_role(): Retrieves IAM Role for attaching to Lambda
        get_policy_attachment(): Attaches IAM Policy to IAM Role
    """

    ecr_repository: aws.ecr.Repository
    lambda_function: aws.lambda_.Function
    image: Image

    def __init__(
        self,
        project_path: str,
        docker_file_path: str,
        resource_prefix: str,
        s3_bucket_name: str,
        ecr_repository_prefix: str,
        lambda_function_memory_size: int,
        lambda_function_timeout: int,
        loki_user: str,
        loki_host: str,
        loki_api_key: str,
        tags: dict,
        env_vars: Dict[str, str],
    ):
        self.project_path = project_path
        self.docker_file_path = docker_file_path
        self.ecr_repository_prefix = ecr_repository_prefix
        # The name of already created s3 bucket
        self.s3_bucket_name = s3_bucket_name
        self.resource_prefix = resource_prefix
        self.loki_user = loki_user
        self.loki_host = loki_host
        self.loki_api_key = loki_api_key

        if not tags:
            tags = {}

        self.tags = tags
        self.env_vars = env_vars

        self.lambda_function_memory_size = lambda_function_memory_size
        self.lambda_function_timeout = lambda_function_timeout

    def deploy(self):
        self.create_ecr_repository()
        self.create_ecr_repository_image()
        self.create_lambda_function()

    def create_ecr_repository(self):
        self.ecr_repository = aws.ecr.Repository(
            self.ecr_repository_prefix,
            tags={"Name": self.ecr_repository_prefix, **self.tags},
        )

    def create_ecr_repository_image(self):
        image_name = self.ecr_repository.repository_url
        registry_info = self.ecr_repository.registry_id.apply(self.__get_registry_info)

        self.image = Image(
            f"{self.resource_prefix}-image",
            build=DockerBuildArgs(
                context=self.project_path,
                dockerfile=self.docker_file_path,
                platform="linux/amd64",
            ),
            image_name=image_name.apply(lambda x: f"{x}:{self.resource_prefix}"),
            registry=registry_info,
        )

    def create_lambda_function(self):
        iam_role = self.get_lambda_iam_role()
        iam_policy = self.get_lambda_iam_policy()

        role_policy_attachment = aws.iam.RolePolicyAttachment(
            f"{self.resource_prefix[:15]}-policy-attachment",
            role=iam_role.name,
            policy_arn=iam_policy.arn,
        )

        # TODO: use lambda extension or other way to paste the environment variables into AWS Lambda
        self.lambda_function = aws.lambda_.Function(
            self.resource_prefix,
            description="Lambda function for execution of PyTest tests in parallel",
            package_type="Image",
            image_uri=self.image.image_name,
            role=iam_role.arn,
            memory_size=self.lambda_function_memory_size,
            timeout=self.lambda_function_timeout,
            environment=aws.lambda_.FunctionEnvironmentArgs(
                variables={
                    "S3_BUCKET": self.s3_bucket_name,
                    "LOKI_USER": self.loki_user,
                    "LOKI_HOST": self.loki_host,
                    "LOKI_API_KEY": self.loki_api_key,
                    **self.env_vars,
                },
            ),
            tags=self.tags,
            opts=pulumi.ResourceOptions(
                depends_on=[
                    self.ecr_repository,
                    self.image,
                    role_policy_attachment,
                ],
            ),
        )

    def get_lambda_iam_policy(self) -> aws.iam.Policy:
        return aws.iam.Policy(
            resource_name=f"{self.resource_prefix[:15]}-policy",
            policy=json.dumps(
                {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Effect": "Allow",
                            "Action": [
                                "s3:PutObject",
                            ],
                            "Resource": f"arn:aws:s3:::{self.s3_bucket_name}/*",
                        },
                        {
                            "Effect": "Allow",
                            "Action": [
                                "logs:CreateLogGroup",
                                "logs:CreateLogStream",
                                "logs:PutLogEvents",
                                "logs:PutRetentionPolicy",
                                "logs:DescribeLogStreams",
                            ],
                            "Resource": [
                                "arn:aws:logs:*:*:*",
                            ],
                        },
                    ],
                }
            ),
        )

    def get_lambda_iam_role(self) -> aws.iam.Role:
        return aws.iam.Role(
            resource_name=f"{self.resource_prefix[:15]}-role",
            assume_role_policy=json.dumps(
                {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Effect": "Allow",
                            "Principal": {"Service": "lambda.amazonaws.com"},
                            "Action": "sts:AssumeRole",
                        }
                    ],
                }
            ),
        )

    @staticmethod
    def __get_registry_info(registry_id) -> RegistryArgs:
        """
        Generates authentication information to access the repository to build and publish the image.
        :param registry_id: Registry ID
        :return: pulumi_docker.ImageRegistry
        """
        credentials = aws.ecr.get_credentials(registry_id=registry_id)
        decoded = base64.b64decode(credentials.authorization_token).decode()

        parts = decoded.split(":")
        if len(parts) != 2:
            raise Exception("Invalid credentials")

        return RegistryArgs(
            server=credentials.proxy_endpoint, username=parts[0], password=parts[1]
        )



from blazetest.core.infra.tools import InfraSetupTool, PulumiInfraSetup
from blazetest.core.utils.exceptions import UnsupportedInfraSetupTool


SUPPORTED_INFRA_SETUP_TOOLS = {
    "pulumi": PulumiInfraSetup,
}


class InfraSetup:
    """
    Infrastructure setup class, used to deploy the artifacts to
    the cloud provider (currently, only AWS is supported)
    """

    def __init__(self, setup_tool: str = "pulumi", *args, **kwargs):
        if setup_tool not in SUPPORTED_INFRA_SETUP_TOOLS:
            raise UnsupportedInfraSetupTool(
                f"{setup_tool} is not supported for deploying, "
                f"supported: {','.join(list(SUPPORTED_INFRA_SETUP_TOOLS.keys()))}"
            )
        self.infra: InfraSetupTool = SUPPORTED_INFRA_SETUP_TOOLS[setup_tool](
            *args, **kwargs
        )

    def deploy(self):
        """
        Deploys the given infrastructure to the cloud provider using given setup tool
        """
        return self.infra.deploy()



from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
import logging
import time
from itertools import repeat
from typing import List, Optional

from blazetest.core.config import MAX_LAMBDA_WORKERS
from blazetest.core.project_config.project_config import BlazetestConfig
from blazetest.core.lambda_.invocation import Invocation
from blazetest.core.run_test.pytest_collect import collect_tests
from blazetest.core.run_test.result_model import (
    TestSessionResult,
    InvocationResult,
    JUnitXMLReport,
    ReportMergeResult,
    TestSessionResultManager,
)
from blazetest.core.utils.exceptions import (
    ReportNotAvailable,
    ReportNotUploaded,
    NoTestsToRun,
    ReportNotMerged,
)
from blazetest.core.utils.logging_config import ColoredOutput
from blazetest.core.utils.report_merger import ReportMerger
from blazetest.core.utils.utils import (
    remove_junit_report_path,
    FILTER_ALL,
    FILTER_FAILED,
    FILTER_FLAKY,
)

logger = logging.getLogger(__name__)


class TestRunner:
    """
    Class for running tests using Lambda functions.
    This class is responsible for collecting the test items,
    creating the report paths, and invoking the Lambda functions.

    Attributes:
        config (BlazetestConfig): Configuration for running the tests on AWS Lambda.
        lambda_function (Lambda):  Object responsible for interacting with AWS Lambda.

    Methods:
        run_tests(): Collects the test items, creates the report paths,
            and invokes the Lambda functions.
    """

    # Lambda function name, which will be invoked to run tests
    function_name: str = None

    # S3 bucket name, where the reports should be saved
    s3_bucket: str = None

    # Retry results from each retry attempt
    retry_test_results: List[InvocationResult] = []

    # (needed for rerun) Manager of the test results that is used to get the function name from the previous run
    tests_result_manager: TestSessionResultManager = None

    # List of test node IDs to be run
    node_ids: List[str] = []

    # Needed for aligning the log output
    longest_node_id_length: int = 0

    def __init__(
        self,
        config: BlazetestConfig,
        uuid: str,
        aws_access_key_id: str = None,
        aws_secret_access_key: str = None,
    ):
        """
        Initializes the class with the given config and creates a new Lambda Invocation and Report Merger objects.
        After, it collects tests and if nothing found, raises NoTestsToRun exception.

        :param config: Configuration for running the tests on AWS Lambda.
        :param uuid: Blazetest session UUID.
        :param test_session: TestSessionResult object from the previous run (used only for rerun).
        :param tests_filter: Filter for the tests to be rerun (used only for rerun). Can be: all | failed.
        """
        self.config = config
        self.uuid = uuid

        resource_prefix = self.config.cloud.aws.get_resource_name(uuid=self.uuid)

        self.lambda_function = Invocation(
            region=self.config.cloud.aws.region,
            resource_prefix=resource_prefix,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )
        self.report_merger = ReportMerger(
            region=self.config.cloud.aws.region,
            resource_prefix=resource_prefix,
            config=config,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )
        self.timestamp = self.__get_timestamp_now()

    def collect_tests(
        self, test_session: TestSessionResult = None, test_filter: str = FILTER_ALL
    ):
        if test_filter == FILTER_ALL:
            self.node_ids = self.get_collected_tests()
        elif test_filter == FILTER_FAILED:
            self.node_ids = test_session.failed_tests_ids
        elif test_filter == FILTER_FLAKY:
            self.node_ids = test_session.flaky_tests_ids

        if self.node_ids:
            logger.info(f"Found {len(self.node_ids)} tests to run")
        else:
            raise NoTestsToRun("Ending session as there are no tests to run")

        self.longest_node_id_length = max(len(node_id) for node_id in self.node_ids)

    def run_tests(
        self,
        flaky_test_retry_enabled: bool = True,
        rerun: bool = False,
    ) -> Optional[TestSessionResult]:
        """
        Creates the report paths, and invokes the Lambda functions in parallel.
        """
        self.set_function_name(rerun=rerun)

        logger.info(
            f"Lambda function: {self.function_name}, " f"S3 bucket: {self.s3_bucket}"
        )
        logger.info("Invoking tests and running in parallel..")

        invocation_result: InvocationResult = self.__run_in_parallel(
            node_ids=self.node_ids
        )

        reports = invocation_result.junit_xml_reports_paths

        if invocation_result.failed_tests_count > 0:
            self.retry(
                flaky_test_retry_enabled=flaky_test_retry_enabled,
                failed_tests=invocation_result.failed_tests_node_ids,
            )

        for retry_invocation_result in self.retry_test_results:
            reports += retry_invocation_result.junit_xml_reports_paths

        report_merge_result = self.__merge_reports(reports=reports)

        return TestSessionResult(
            uuid=self.uuid,
            lambda_function_name=self.function_name,
            tests_count=len(self.node_ids),
            tests_passed=report_merge_result.passed,
            failed_tests_count=invocation_result.failed_tests_count,
            failed_tests_ids=report_merge_result.failed_ids,
            flaky_tests_count=report_merge_result.flaky,
            flaky_tests_ids=report_merge_result.flaky_ids,
            flake_detected=self.report_merger.flake_detected,
            pytest_duration=invocation_result.pytest_duration,
            s3_bucket=self.s3_bucket,
            start_timestamp=self.timestamp,
            end_timestamp=self.__get_timestamp_now(),
            junit_report_path=report_merge_result.final_report_path,
            flake_report_path=report_merge_result.flake_report_path
            if self.report_merger.flake_detected
            else None,
            config=self.config,
        )

    def get_collected_tests(self):
        return collect_tests(pytest_args=self.config.framework.pytest.collection_args)

    def retry(
        self,
        flaky_test_retry_enabled: bool,
        failed_tests: List[str],
    ):
        if flaky_test_retry_enabled and failed_tests:
            logger.info(
                f"Retrying running {len(failed_tests)} failed tests "
                f"{self.config.general.flaky.failure_retry_attempts} times"
            )
            # saves the result of each retry in self.retry_test_results
            self.__retry_failed_tests(failed_tests=failed_tests)

    def __retry_failed_tests(self, failed_tests: List[str], retry_num: int = 0):
        if retry_num >= self.config.general.flaky.failure_retry_attempts:
            return

        retry_test_results = self.__run_in_parallel(
            node_ids=failed_tests,
            retry=True,
        )
        self.retry_test_results.append(retry_test_results)

        logger.info(
            f"Retry #{retry_num + 1}, failed: {retry_test_results.failed_tests_count} "
            f"out of {retry_test_results.passed_tests_count + retry_test_results.failed_tests_count}"
        )

        if (
            retry_test_results.failed_tests_count == 0
            and self.config.general.flaky.exit_on_flake_detection is True
        ):
            logger.info(
                f"All failed tests passed on retry attempt: {retry_num + 1}. Stopping retrying."
            )
            return

        self.__retry_failed_tests(
            failed_tests=(
                retry_test_results.failed_tests_node_ids
                if self.config.general.flaky.exit_on_flake_detection is True
                else failed_tests
            ),
            retry_num=retry_num + 1,
        )

    def __run_in_parallel(self, node_ids: List[str], retry: bool = False):
        start_time = time.time()

        print(
            f"* Observing {len(node_ids)} tests completion status "
            f"{ColoredOutput.GREEN.value}...{ColoredOutput.RESET.value}"
        )
        # TODO: is 1000 appropriate value for workers for threads? Move to async?
        with ThreadPoolExecutor(max_workers=MAX_LAMBDA_WORKERS) as executor:
            results = list(executor.map(self.invoke_lambda, node_ids, repeat(retry)))

        return InvocationResult.parse(results=results, start_time=start_time)

    def invoke_lambda(self, node_id: str, retry: bool = False):
        print(f"* [{node_id}] ... Running")

        report_path = self.__get_pytest_xml_report_path(node_id=node_id)
        pytest_args = remove_junit_report_path(
            self.config.framework.pytest.execution_args
        )

        invocation_result = self.lambda_function.invoke(
            function_name=self.function_name,
            session_uuid=self.uuid,
            node_id=node_id,
            pytest_args=pytest_args,
            report_path=report_path,
            timestamp=self.timestamp,
            retry=retry,
        )

        if invocation_result["test_result"]:
            print(
                f"* [{node_id:{self.longest_node_id_length}}] ... Finished "
                f"{ColoredOutput.GREEN.value}[passed]{ColoredOutput.RESET.value}"
            )
        else:
            print(
                f"* [{node_id:{self.longest_node_id_length}}] ... Finished "
                f"{ColoredOutput.RED.value}[failed]{ColoredOutput.RESET.value}"
            )

        return node_id, invocation_result

    def set_function_name(self, rerun: bool = False):
        if not rerun:
            self.function_name = (
                self.lambda_function.get_created_lambda_function_details()
            )
            return

        test_session = self.tests_result_manager.get_test_session_by_uuid(
            uuid=self.uuid,
        )

        if test_session is None:
            logger.error(f"No test session with the given UUID found: {self.uuid}")
            return None

        self.function_name = test_session.lambda_function_name
        self.s3_bucket = test_session.s3_bucket

    def __merge_reports(self, reports: List[JUnitXMLReport]) -> ReportMergeResult:
        """
        Merges the reports saved to s3 bucket using junitparser and saves the final result in the s3 bucket.

        :param reports: list of reports to be merged

        Returns the path to the merged report.
        """
        self.report_merger.set_s3_bucket_name(s3_bucket_name=self.s3_bucket)

        try:
            report_merge_result: ReportMergeResult = self.report_merger.merge_reports(
                reports=reports,
                timestamp=self.timestamp,
            )
        except (ReportNotAvailable, ReportNotUploaded) as e:
            raise ReportNotMerged(
                f"Error merging reports: {str(e)}. Try rerunning the session."
            )

        logger.info(
            f"Report merged and saved to s3://{self.s3_bucket}/{report_merge_result.final_report_path}"
        )
        return report_merge_result

    def set_s3_bucket_name(self, s3_bucket_name: str):
        self.s3_bucket = s3_bucket_name

    def set_tests_result_manager(self, tests_result_manager: TestSessionResultManager):
        self.tests_result_manager = tests_result_manager

    REPLACE_SYMBOLS = ["::", ".", "/"]
    REPLACE_TO = "-"
    TMP_REPORT_FOLDER = "/tmp/junitxml/{}.xml"
    FOLDER_NAME_TIMESTAMP = "%Y-%m-%d_%H-%M-%S"

    def __get_timestamp_now(self):
        return datetime.now().strftime(self.FOLDER_NAME_TIMESTAMP)

    def __get_pytest_xml_report_path(self, node_id: str) -> str:
        for symbol in self.REPLACE_SYMBOLS:
            node_id = node_id.replace(symbol, self.REPLACE_TO)

        return self.TMP_REPORT_FOLDER.format(node_id)



import logging

from blazetest.core.license.manager import LicenseManager
from blazetest.core.project_config.model import BlazetestConfig
from blazetest.core.utils.exceptions import (
    LicenseNotValid,
    LicenseExpired,
    LicenseNotSpecified,
)

logger = logging.getLogger(__name__)


def check_license(license_key, license_file, config: BlazetestConfig) -> tuple:
    try:
        licence_manager = LicenseManager(
            license_key=license_key,
            license_file=license_file,
            config_license_key=config.general.license.license_key,
            config_license_file=config.general.license.license_file,
        )
        expiration_date = licence_manager.check_license()
    except (LicenseNotSpecified, LicenseExpired, LicenseNotValid) as e:
        logger.error(e)
        return None, None

    logger.info(f"License expiration date: {expiration_date}")

    flaky_enabled = False
    if licence_manager.flaky_test_retry_enabled():
        logger.info("Flaky test retry feature is on")
        if config.general.flaky.failure_retry_attempts > 0:
            flaky_enabled = True
    else:
        if config.general.flaky.failure_retry_attempts > 0:
            logger.warning(
                "Flaky test retry feature is not available within this license"
            )

    return expiration_date, flaky_enabled


EXIT_SUCCESS = 1
EXIT_FAILURE = 0



from abc import ABC, abstractmethod
import logging
from pulumi import automation as auto
from pulumi.automation import LocalWorkspaceOptions, ProjectBackend, ProjectSettings

from blazetest.core.config import CWD, LOKI_HOST, LOKI_USER, DOCKER_FILE_PATH
from blazetest.core.deployment.aws import AWSWorkflow

logger = logging.getLogger(__name__)


class InfraSetupTool(ABC):
    def __init__(
        self,
        aws_region: str,
        resource_prefix: str,
        s3_bucket_name: str,
        ecr_repository_prefix: str,
        lambda_function_timeout: int,
        lambda_function_memory_size: int,
        loki_api_key: str,
        tags: dict,
        debug: bool,
    ):
        self.aws_region = aws_region
        self.resource_prefix = resource_prefix
        self.s3_bucket_name = s3_bucket_name
        self.ecr_repository_prefix = ecr_repository_prefix
        self.lambda_function_timeout = lambda_function_timeout
        self.lambda_function_memory_size = lambda_function_memory_size
        self.loki_api_key = loki_api_key
        self.tags = tags
        self.debug = debug

    @abstractmethod
    def deploy(self) -> None:
        pass


def log_pulumi_event(event: str):
    logger.info(event)


class PulumiInfraSetup(InfraSetupTool):
    """
    Uses Pulumi (https://www.pulumi.com/docs/) Automation API to build and deploy artifacts to the cloud.
    """

    PROJECT_NAME = "blazetest"
    PROJECT_BACKEND_URL = "file:\\/\\/~"
    PROJECT_RUNTIME = "python"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def deploy(self) -> None:
        # TODO: User should pass env variable himself
        env_vars = {}

        workflow = AWSWorkflow(
            project_path=CWD,
            docker_file_path=DOCKER_FILE_PATH,
            resource_prefix=self.resource_prefix,
            s3_bucket_name=self.s3_bucket_name,
            ecr_repository_prefix=self.ecr_repository_prefix,
            lambda_function_timeout=self.lambda_function_timeout,
            lambda_function_memory_size=self.lambda_function_memory_size,
            loki_host=LOKI_HOST,
            loki_user=LOKI_USER,
            loki_api_key=self.loki_api_key,
            env_vars=env_vars,
            tags=self.tags,
        )

        stack = auto.create_stack(
            stack_name=self.resource_prefix,
            project_name="blazetest",
            program=workflow.deploy,
            opts=self.get_project_settings(),
        )

        logger.info("Installing plugins")

        # TODO: updated automatically to the stable version
        stack.workspace.install_plugin("aws", "v5.42.0")
        stack.workspace.install_plugin("docker", "v4.3.1")

        stack.set_config("aws:region", auto.ConfigValue(value=self.aws_region))
        stack.refresh(on_output=log_pulumi_event, show_secrets=False)

        logger.info("Deploying..")
        workflow_result = stack.up(  # noqa
            show_secrets=False, on_output=log_pulumi_event, debug=self.debug
        )

        logger.info(
            "Deploying has finished.",
        )

    def get_project_settings(self):
        return LocalWorkspaceOptions(
            project_settings=ProjectSettings(
                name=self.PROJECT_NAME,
                backend=ProjectBackend(url=self.PROJECT_BACKEND_URL),
                runtime=self.PROJECT_RUNTIME,
            ),
        )



import logging
from typing import Dict

import toml
from dacite import from_dict, MissingValueError, WrongTypeError

from blazetest.core.config import PROJECT_CONFIG_TOML
from blazetest.core.project_config.model import BlazetestConfig
from blazetest.core.utils.exceptions import (
    ConfigurationFileNotFound,
    ConfigurationMissingValue,
    ConfigurationFieldWrongType,
)

logger = logging.getLogger(__name__)


class ProjectConfiguration:
    @classmethod
    def from_toml_file(cls, toml_file_path: str = None) -> BlazetestConfig:
        """Load configuration data from a TOML file and create a `BlazetestConfig` object.

        Args:
            toml_file_path: Path to the TOML file. If not specified, the default
                location specified by `BLAZETEST_CONFIG` will be used.

        Returns:
            A `BlazetestConfig` object created from the data in the TOML file.

        Raises:
            ConfigurationFileNotFound: If the specified TOML file does not exist.
            ConfigurationMissingValue: If required values are missing from the TOML file.
            ConfigurationFieldWrongType: If a value in the TOML file has the wrong type.
        """
        if not toml_file_path:
            logger.info(
                f"Config file location not specified. Using default location: {PROJECT_CONFIG_TOML}",
            )
            toml_file_path = PROJECT_CONFIG_TOML

        try:
            with open(toml_file_path) as f:
                data = toml.loads(f.read())
                return cls.get_dataclass_from_dict(data=data)
        except FileNotFoundError as err:
            raise ConfigurationFileNotFound(f"Configuration file does not exist: {err}")

    @classmethod
    def from_toml_string(cls, toml_string: str) -> BlazetestConfig:
        """Load configuration data from a TOML string and create a `LambdaConfig` object.

        Args:
            toml_string: TOML string containing configuration data.

        Returns:
            A `LambdaConfig` object created from the data in the TOML string.

        Raises:
            ConfigurationMissingValue: If required values are missing from the TOML string.
            ConfigurationFieldWrongType: If a value in the TOML string has the wrong type.
        """
        try:
            data = toml.loads(toml_string)
            return cls.get_dataclass_from_dict(data=data)
        except FileNotFoundError as err:
            raise ConfigurationFileNotFound(f"Configuration file does not exist: {err}")

    @staticmethod
    def get_dataclass_from_dict(data: Dict):
        """Create a `LambdaConfig` object from a dictionary.

        Args:
            data: Dictionary containing configuration data.

        Returns:
            A `LambdaConfig` object created from the data in the dictionary.

        Raises:
            ConfigurationMissingValue: If required values are missing from the dictionary.
            ConfigurationFieldWrongType: If a value in the dictionary has the wrong type.
        """
        try:
            return from_dict(data_class=BlazetestConfig, data=data)
        except MissingValueError as err:
            raise ConfigurationMissingValue(f"Blazetest configuration error: {err}")
        except WrongTypeError as err:
            raise ConfigurationFieldWrongType(f"Blazetest configuration error: {err}")



import logging
from typing import Optional
import sys

import click
from art import text2art

from blazetest.cli.base import cli
from blazetest.cli.utils import check_license, EXIT_SUCCESS, EXIT_FAILURE
from blazetest.core.infra.infra_setup import InfraSetup
from blazetest.core.project_config.model import BlazetestConfig
from blazetest.core.project_config.project_config import ProjectConfiguration
from blazetest.core.run_test.runner_facade import TestRunner
from blazetest.core.run_test.result_model import (
    TestSessionResult,
    TestSessionResultManager,
)
from blazetest.core.utils.exceptions import NoTestsToRun, ReportNotMerged
from blazetest.core.utils.logging_config import setup_logging
from blazetest.core.utils.s3_manager import S3Manager
from blazetest.core.utils.utils import (
    create_build_folder,
    generate_uuid,
    set_environment_variables,
    combine_tags,
    LOGS_ENABLED,
    LOGS_DISABLED,
    SETUP_TOOL_PULUMI,
)

logger = logging.getLogger(__name__)


@cli.command()
@click.option(
    "-config",
    "--config-path",
    help="Configuration path to the TOML file for the CLI. "
    "If not specified -> searches the project's root folder for the file blazetest.toml. "
    "If not found -> raises an error.",
)
@click.option(
    "-ak",
    "--aws-access-key-id",
    help="AWS Access Key ID which is used to deploy artifacts.",
)
@click.option(
    "-as",
    "--aws-secret-access-key",
    help="AWS Secret Access Key which is used to deploy artifacts.",
)
@click.option(
    "-k",
    "--license-key",
    help="License key for Blazetest CLI. Either --license-key or --license-file should be specified."
    "License key has a higher priority if both are specified.",
)
@click.option(
    "-l",
    "--license-file",
    help="License file for Blazetest CLI. Either --license-key or --license-file should be specified."
    "License file has a lower priority if both are specified.",
)
@click.option(
    "-t",
    "--tags",
    help="Tags specified for the AWS Lambda function. The tags will be attached to created Lambda function instance.",
)
@click.option(
    "-lo",
    "--logs",
    default=LOGS_ENABLED,
    type=click.Choice([LOGS_ENABLED, LOGS_DISABLED]),
    help="Default is enabled. When logs are enabled, they are shown in the console stdout. "
    "When they are set to disabled, the logs are not shown during CLI execution, but saved to blazetest.log, "
    "which will be located in the project's root.",
)
@click.option(
    "-rl",
    "--remote-logs",
    help="If provided, logs are sent to the remote logging system. Default is Loki",
)
@click.option("-de", "--debug", is_flag=True, help="Enables debugging output.")
def run(
    config_path: str,
    aws_access_key_id: str,
    aws_secret_access_key: str,
    license_key: str,
    license_file: str,
    tags: str,
    logs: str,
    remote_logs: str,
    debug: bool,
):
    """
    Runs tests using the PyTest library and parallel Lambda functions.
    It deploys the necessary AWS resources (ECR, Lambda, and S3 bucket) using Pulumi.
    """
    print(text2art("BlazeTest"))
    session_uuid = generate_uuid()
    setup_logging(
        debug=debug,
        stdout_enabled=logs != LOGS_DISABLED,
        loki_api_key=remote_logs,
        session_uuid=session_uuid,
    )
    logger.info(f"Session UUID: {session_uuid}")

    # Get project configuration from Blazetest TOML
    config = ProjectConfiguration.from_toml_file(config_path)
    expiration_date, flaky_enabled = check_license(license_key, license_file, config)
    if expiration_date is None:
        return

    test_runner = TestRunner(
        config=config,
        uuid=session_uuid,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )

    try:
        test_runner.collect_tests()
    except NoTestsToRun as e:
        logger.error(e)
        return

    set_environment_variables(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )

    all_tags = combine_tags(
        tags=tags, config_tags=config.cloud.aws.tags, session_uuid=session_uuid
    )

    s3_manager = S3Manager(
        aws_config=config.cloud.aws,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )
    s3_bucket_name = s3_manager.find_or_create_s3_bucket(tags=all_tags)

    deploy(
        config=config,
        s3_bucket_name=s3_bucket_name,
        session_uuid=session_uuid,
        tags=all_tags,
        debug=debug,
        remote_logging_api_key=remote_logs,
    )

    test_session_result_manager = TestSessionResultManager(
        config=config,
        s3_manager=s3_manager,
    )

    test_runner.set_tests_result_manager(test_session_result_manager)
    test_runner.set_s3_bucket_name(s3_bucket_name)

    try:
        test_session_result: Optional[TestSessionResult] = test_runner.run_tests(
            flaky_test_retry_enabled=flaky_enabled,
        )
    except ReportNotMerged as err:
        logger.error(err)
        sys.exit(EXIT_FAILURE)

    if test_session_result is not None:
        test_session_result.log_results(failed_test_retry_enabled=flaky_enabled)
        test_session_result.set_tags(tags if tags else None)
        test_session_result.set_uuid(session_uuid)
        test_session_result_manager.append_results_to_json(test_session_result)

        if test_session_result.failed_tests_count > 0:
            sys.exit(EXIT_FAILURE)

        if (
            config.general.flaky.fail_on_flake is True
            and test_session_result.flake_detected
        ):
            sys.exit(EXIT_FAILURE)

        sys.exit(EXIT_SUCCESS)

    sys.exit(EXIT_SUCCESS)


def deploy(
    config: BlazetestConfig,
    s3_bucket_name: str,
    session_uuid: str,
    tags: dict,
    remote_logging_api_key: str,
    debug: bool,
):
    """
    Creates the build folder and deploys the necessary AWS resources (ECR, Lambda, and S3 bucket) using Pulumi.
    """
    remote_logging_enabled = remote_logging_api_key is not None
    create_build_folder(remote_logging_enabled=remote_logging_enabled)
    infra_setup = InfraSetup(
        aws_region=config.cloud.aws.region,
        resource_prefix=config.cloud.aws.get_resource_name(uuid=session_uuid),
        ecr_repository_prefix=config.cloud.aws.get_ecr_repository_prefix(
            uuid=session_uuid
        ),
        lambda_function_memory_size=config.cloud.aws.lambda_function_memory_size,
        lambda_function_timeout=config.cloud.aws.lambda_function_timeout,
        s3_bucket_name=s3_bucket_name,
        tags=tags if tags else None,
        loki_api_key=remote_logging_api_key,
        setup_tool=SETUP_TOOL_PULUMI,
        debug=debug,
    )
    infra_setup.deploy()



import json
import logging
import os
from typing import List, Dict

import boto3
import pytest

PWD = os.path.dirname(__file__)
logger = logging.getLogger(__name__)


def setup_logging():
    logging.basicConfig(
        format="%(process)d-%(levelname)s-%(message)s", level=logging.INFO
    )


class S3Upload:
    def __init__(self, region):
        self.client = boto3.client(
            "s3",
            region_name=region,
        )

    def upload_file_to_s3_bucket(
        self,
        filepath: str,
        timestamp: str,
        s3_bucket: str,
        session_uuid: str,
        retry: bool,
    ) -> str:
        filename = os.path.basename(filepath)
        dst_folder = os.path.join(session_uuid, f"{timestamp}/target/junitxml")
        dst_filepath = os.path.join(dst_folder, filename)

        if retry:
            dst_filepath = os.path.join(dst_folder, f"flaky-{filename}")

        with open(filepath, "rb") as f:
            self.client.put_object(
                Body=f,
                Bucket=s3_bucket,
                Key=dst_filepath,
            )

        return dst_filepath


def execute_tests(args: List[str]):
    return pytest.main(args)


def run_tests(event, context=None) -> Dict:  # noqa
    setup_logging()

    pytest_args: List[str] = event["pytest_args"]
    node_id: str = event["node_id"]
    report_path: str = event["report_path"]
    region: str = event["region"]
    session_uuid: str = event["session_uuid"]
    timestamp: str = event["start_timestamp"]
    retry: bool = event["retry"]

    logger.info(f"Invoking test: {node_id} with pytest args: {pytest_args}")

    s3_bucket = os.environ.get("S3_BUCKET")

    args = [node_id, f"--junitxml={report_path}"] + pytest_args

    exit_code = execute_tests(args=args)
    test_result = True if exit_code == 0 else False

    logger.info(f"Test result: {'passed' if test_result else 'failed'}")

    s3 = S3Upload(region=region)
    s3_path = s3.upload_file_to_s3_bucket(
        filepath=report_path,
        session_uuid=session_uuid,
        s3_bucket=s3_bucket,
        timestamp=timestamp,
        retry=retry,
    )

    test_result = {
        "test_result": test_result,
        "report_path": s3_path,
    }

    response = {
        "statusCode": 200,
        "body": json.dumps(test_result),
    }

    logger.info(f"Response: {response}")

    return response



import logging
import os
import platform
import shutil
import uuid
from pathlib import Path
from typing import List, Any, Optional

from blazetest.core.config import (
    BUILD_FOLDER_PATH,
    PWD,
    EXECUTABLE_FILES,
)

logger = logging.getLogger(__name__)

FILTER_ALL = "all"
FILTER_FAILED = "failed"
FILTER_FLAKY = "flaky"

LOGS_ENABLED = "enabled"
LOGS_DISABLED = "disabled"

SETUP_TOOL_PULUMI = "pulumi"


def parse_tags(tags: str = None) -> dict:
    """
    Parses tags from CLI in the format of 'key1=value1,key2=value2' into a dictionary

    :param tags: string of tags from CLI in the format of 'key1=value1,key2=value2'
    :return: dictionary of tags in the format of {'key1': 'value1', 'key2': 'value2'}
    """
    tags_d = {}

    if tags:
        tag_values = tags.split(",")

        for tag_value in tag_values:
            tag = tag_value.split("=")

            if len(tag) != 2:
                continue

            tags_d[tag[0]] = tag[1]

    return tags_d


def combine_tags(
    config_tags: Any, session_uuid: str, tags: Optional[str] = None
) -> dict:
    """
    Combines tags from CLI, config file and adds session UUID as a tag into one dictionary

    :param tags: string of tags from CLI in the format of 'key1=value1,key2=value2'
    :param config_tags: tags from BlazetestConfig object
    :param session_uuid: string of session UUID
    :return: dictionary with all tags in one dictionary
    """
    tags = parse_tags(tags)
    tags.update(dict(config_tags))
    tags["blazetest:uuid"] = session_uuid
    return tags


def generate_uuid() -> str:
    return str(uuid.uuid4())[:8]


def get_python_version() -> str:
    """
    Gets current Python version

    :return: string in the format of: '3.8', '3.7'
    """
    return ".".join(list(platform.python_version_tuple()[:2]))


def remove_junit_report_path(pytest_args: List[str]) -> List[str]:
    """Remove the `--junitxml` argument from a list of pytest arguments.

    :param pytest_args: List of pytest arguments.
    :return: List of pytest arguments with the `--junitxml` argument removed.
    """
    return [arg for arg in pytest_args if not arg.startswith("--junitxml")]


def set_environment_variables(
    aws_access_key_id: str = None,
    aws_secret_access_key: str = None,
    aws_region: str = None,
):
    """Set AWS credentials as environment variables.

    :param aws_access_key_id: AWS access key ID.
    :param aws_secret_access_key: AWS secret access key.
    :param aws_region: AWS region.
    """
    if aws_access_key_id:
        os.environ["AWS_ACCESS_KEY_ID"] = aws_access_key_id

    if aws_secret_access_key:
        os.environ["AWS_SECRET_ACCESS_KEY"] = aws_secret_access_key

    if aws_region:
        os.environ["AWS_DEFAULT_REGION"] = aws_region

    os.environ["PULUMI_CONFIG_PASSPHRASE"] = "pass"


def insert_python_version(dockerfile_path, python_version):
    """
    Inserts an indicated Python version to the Dockerfile variable PYTHON_VERSION

    :param dockerfile_path: Path to the Dockerfile
    :param python_version: Python version in the format '3.8'
    """
    with open(dockerfile_path, "r") as f:
        lines = f.read()

    lines = lines.replace("PYTHON_VERSION", python_version)
    with open(dockerfile_path, "w") as f:
        f.write(lines)


# TODO: optimize function to be more readable and more efficient
def create_build_folder(remote_logging_enabled: bool) -> None:
    """
    Create a build folder and copy necessary files to it.

    Files located in .blazetest folder:
    - Dockerfile
    - scripts
        - install_dependencies.sh
    - tests_runner_handler
        - __init__.py
        - tests_runner_handler.py
    """
    # Create build folder and necessary subdirectories (scripts and tests_runner_handler)
    Path(BUILD_FOLDER_PATH).mkdir(parents=True, exist_ok=True)
    Path(os.path.join(BUILD_FOLDER_PATH, "scripts")).mkdir(
        parents=True,
        exist_ok=True,
    )
    Path(os.path.join(BUILD_FOLDER_PATH, "tests_runner_handler")).mkdir(
        parents=True,
        exist_ok=True,
    )

    # Create list of source-destination file pairs for the files in EXECUTABLE_FILES
    src_dst_pairs = [
        (os.path.join(PWD, file), os.path.join(BUILD_FOLDER_PATH, file))
        for file in EXECUTABLE_FILES
    ]

    for src, dst in src_dst_pairs:
        shutil.copyfile(src=src, dst=dst)

    dockerfile = "RemoteLogging.Dockerfile" if remote_logging_enabled else "Dockerfile"

    shutil.copyfile(
        src=os.path.join(PWD, dockerfile),
        dst=os.path.join(BUILD_FOLDER_PATH, "Dockerfile"),
    )

    insert_python_version(
        dockerfile_path=os.path.join(BUILD_FOLDER_PATH, "Dockerfile"),
        python_version=get_python_version(),
    )

    logger.info("Successfully created build folder")


def print_table(data):
    max_lengths = [max(map(len, col)) for col in zip(*data)]

    for row in data:
        print("   ".join(word.ljust(length) for word, length in zip(row, max_lengths)))



import logging
import sys
from typing import Optional

import click
from art import text2art

from blazetest.cli.base import cli
from blazetest.cli.utils import check_license, EXIT_FAILURE, EXIT_SUCCESS
from blazetest.core.project_config.project_config import ProjectConfiguration
from blazetest.core.run_test.runner_facade import TestRunner
from blazetest.core.run_test.result_model import (
    TestSessionResult,
    TestSessionResultManager,
)
from blazetest.core.utils.exceptions import NoTestsToRun
from blazetest.core.utils.logging_config import setup_logging
from blazetest.core.utils.s3_manager import S3Manager
from blazetest.core.utils.utils import (
    generate_uuid,
    set_environment_variables,
    FILTER_ALL,
    FILTER_FAILED,
    FILTER_FLAKY,
)

logger = logging.getLogger(__name__)


@cli.command()
@click.option(
    "-ri",
    "--runid",
    help="Blazetest session UUID to rerun.",
    required=True,
)
@click.option(
    "-config",
    "--config-path",
    help="Configuration path to the TOML file for the CLI. "
    "If not specified -> searches the project's root folder for the file blazetest.toml. "
    "If not found -> raises an error.",
)
@click.option(
    "-ak",
    "--aws-access-key-id",
    help="AWS Access Key ID which is used to deploy artifacts.",
)
@click.option(
    "-as",
    "--aws-secret-access-key",
    help="AWS Secret Access Key which is used to deploy artifacts.",
)
@click.option(
    "-k",
    "--license-key",
    help="License key for Blazetest CLI. Either --license-key or --license-file should be specified."
    "License key has a higher priority if both are specified.",
)
@click.option(
    "-l",
    "--license-file",
    help="License file for Blazetest CLI. Either --license-key or --license-file should be specified."
    "License file has a lower priority if both are specified.",
)
@click.option(
    "-lo",
    "--logs",
    default="enabled",
    type=click.Choice(["enabled", "disabled"]),
    help="Default is enabled. When logs are enabled, they are shown in the console stdout. "
    "When they are set to disabled, the logs are not shown during CLI execution, but saved to blazetest.log, "
    "which will be located in the project's root.",
)
@click.option(
    "-lk",
    "--loki",
    help="Loki API Key. If provided, logs are sent to the Loki",
)
@click.option(
    "-s",
    "--status",
    default=FILTER_ALL,
    type=click.Choice([FILTER_ALL, FILTER_FAILED, FILTER_FLAKY]),
    help="Filter which kind of tests to rerun (all, failed)",
)
@click.option("-de", "--debug", is_flag=True, help="Enables debugging output.")
def rerun(
    runid: str,
    config_path: str,
    aws_access_key_id: str,
    aws_secret_access_key: str,
    license_key: str,
    license_file: str,
    logs: str,
    loki: str,
    status: str,
    debug: bool,
):
    """
    Reruns the session specified with --uuid option.
    """
    print(text2art("BlazeTest"))
    rerun_uuid = generate_uuid()

    setup_logging(
        debug=debug,
        stdout_enabled=logs != "disabled",
        loki_api_key=loki,
        session_uuid=rerun_uuid,
    )

    config = ProjectConfiguration.from_toml_file(config_path)
    expiration_date, flaky_enabled = check_license(license_key, license_file, config)
    if expiration_date is None:
        return

    logger.info(f"Rerunning session with UUID: {runid}")
    logger.info(f"Rerun UUID: {rerun_uuid}")

    s3_manager = S3Manager(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        aws_config=config.cloud.aws,
    )

    s3_bucket_name = s3_manager.find_s3_bucket()
    if s3_bucket_name is None:
        logger.error("S3 Bucket with previous sessions not found")
        return

    test_session_result_manager = TestSessionResultManager(
        config=config,
        s3_manager=s3_manager,
    )

    test_session: TestSessionResult = (
        test_session_result_manager.get_test_session_by_uuid(uuid=runid)
    )
    if test_session is None:
        logger.error(f"Session with UUID {runid} not found")
        return

    config = ProjectConfiguration.get_dataclass_from_dict(test_session.config)

    test_runner = TestRunner(
        config=config,
        uuid=runid,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )

    try:
        test_runner.collect_tests(
            test_filter=status,
            test_session=test_session,
        )
    except NoTestsToRun as e:
        logger.error(e)
        return

    set_environment_variables(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )

    test_runner.set_s3_bucket_name(s3_bucket_name)
    test_runner.set_tests_result_manager(test_session_result_manager)

    test_session_result: Optional[TestSessionResult] = test_runner.run_tests(
        flaky_test_retry_enabled=flaky_enabled,
        rerun=True,
    )

    if test_session_result is not None:
        test_session_result.log_results(failed_test_retry_enabled=flaky_enabled)
        test_session_result.set_uuid(runid)
        test_session_result.set_rerun_uuid(rerun_uuid)
        test_session_result_manager.append_results_to_json(test_session_result)

        if test_session_result.failed_tests_count > 0:
            sys.exit(EXIT_FAILURE)

        if (
            config.general.flaky.fail_on_flake is True
            and test_session_result.flake_detected
        ):
            sys.exit(EXIT_FAILURE)

        sys.exit(EXIT_SUCCESS)

    sys.exit(EXIT_FAILURE)



import enum
import logging
import os
import uuid

import click
import logging_loki

from blazetest.core.config import LOKI_URL, CWD


class ColoredFormatter(logging.Formatter):
    def format(self, record):
        level = record.levelname
        msg = record.msg
        if level == "DEBUG":
            msg = click.style(msg, fg="blue")
        elif level == "INFO":
            msg = click.style(msg, bold=True)
        elif level == "WARNING":
            msg = click.style(msg, fg="yellow")
        elif level == "ERROR":
            msg = click.style(msg, fg="red", bold=True)
        elif level == "CRITICAL":
            msg = click.style(msg, bg="red", fg="white")
        return f"* {msg}"


def setup_logging(
    debug: bool = False,
    stdout_enabled: bool = True,
    loki_api_key: str = None,
    session_uuid: str = uuid.uuid4(),
):
    """
    Sets up basic logging.
    If stdout_enabled, stdout is shown to the user. Otherwise, saved to the file.
    If loki_api_key is provided, logs are sent to Loki.
    """
    level = logging.DEBUG if debug else logging.INFO

    handlers = []
    # TODO: debug not working well with Loki (possible reason: too many requests)
    if loki_api_key:
        logging_loki.emitter.LokiEmitter.level_tag = "level"
        handler = logging_loki.LokiHandler(
            url=LOKI_URL.format(loki_api_key=loki_api_key),
            tags={"service": "blazetest", "session_id": session_uuid},
            version="1",
        )
        handlers.append(handler)

    if stdout_enabled:
        colored_handler = logging.StreamHandler()
        colored_handler.setFormatter(ColoredFormatter())
        handlers.append(colored_handler)
    else:
        handlers.append(
            logging.FileHandler(filename=os.path.join(CWD, "blazetest.log"))
        )

    logging.basicConfig(
        format="%(message)s",
        level=level,
        handlers=handlers,
    )


class ColoredOutput(enum.Enum):
    RED = "\033[91m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RESET = "\033[0m"



import json
import logging
from typing import List

import boto3
from botocore.config import Config

from blazetest.core.config import MAX_LAMBDA_WORKERS
from blazetest.core.utils.exceptions import LambdaNotCreated

logger = logging.getLogger(__name__)


class Invocation:
    """
    Invocation class, which is used to invoke AWS Lambda

    Attributes:
        region: AWS region.
        resource_prefix: Used CloudFormation stack name

    Methods:
        invoke(): invokes Lambda function with payload
        get_created_lambda_function_details(): returns the exact function name and
            S3 Bucket that it has access to
    """

    LAMBDA_INVOCATION_TYPE = "RequestResponse"
    LAMBDA_LOG_TYPE = "Tail"

    def __init__(
        self,
        region: str = None,
        resource_prefix: str = None,
        aws_access_key_id: str = None,
        aws_secret_access_key: str = None,
    ):
        self.region = region
        self.resource_prefix = resource_prefix
        self.client = boto3.client(
            "lambda",
            region_name=region,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            config=Config(max_pool_connections=MAX_LAMBDA_WORKERS),
        )

    def invoke(
        self,
        node_id: str,
        session_uuid: str,
        function_name: str,
        pytest_args: List[str],
        report_path: str,
        timestamp: str,
        retry: bool = False,
    ) -> dict:
        """Invoke a Lambda function and retrieve the test result.

        Args:
            session_uuid: UUID of the test session.
            node_id: ID of the node to be passed to the Lambda function.
            function_name: Name of the Lambda function to be invoked.
            pytest_args: List of pytest arguments to be passed to the function.
            report_path: Path to the JUnit XML report to be generated by the function.
            timestamp: Timestamp to be passed to the function.
            retry: Boolean value to indicate if the test is a rerun.

        Returns:
            dict: A dictionary containing the test result and the path to the report.
        """
        payload = {
            "pytest_args": pytest_args,
            "node_id": node_id,
            "session_uuid": session_uuid,
            "report_path": report_path,
            "region": self.region,
            "start_timestamp": timestamp,
            "retry": retry,
        }
        return self.process_response(
            response=self.client.invoke(
                FunctionName=function_name,
                InvocationType=self.LAMBDA_INVOCATION_TYPE,
                LogType=self.LAMBDA_LOG_TYPE,
                Payload=json.dumps(payload),
            )
        )

    @staticmethod
    def process_response(response) -> dict:
        """
        Converts response JSON to Python-dict and returns a bool if test passed.
        Raises an error if there is any error message.

        :param response:
        :return:
        """
        result_payload = json.loads(response["Payload"].read())
        try:
            result_body = json.loads(result_payload["body"])
            logger.debug(result_body)
        except KeyError:
            error_message = result_payload["errorMessage"]
            logger.error(
                f"There was an error during test execution: {error_message}",
            )
            return {"test_result": False, "report_path": ""}
        return result_body

    # TODO: is it possible to use other method for retrieving details of created lambda function?
    def get_created_lambda_function_details(self) -> str:
        """
        Retrieves the details of a specific Lambda function that was created.

        The function queries the list of functions from the AWS Lambda service using the `client` object.
        It looks for a function whose name starts with a resource_prefix.
        The function then saves the name of the function in a dictionary.

        Raises:
        If it does not find a function with the specific name, it raises LambdaNotCreated.

        It returns the dictionary with the function details if it is able to find the function
            with the specific namee.

        Returns:
            str : A dictionary containing the function name value.
        """
        functions = []
        response = self.client.list_functions()
        while response:
            functions.extend(response["Functions"])
            response = (
                self.client.list_functions(Marker=response["NextMarker"])
                if "NextMarker" in response
                else None
            )

        function = None
        for f in functions:
            if f["FunctionName"].startswith(f"{self.resource_prefix}"):
                function = f
                break

        if function is None:
            raise LambdaNotCreated("Lambda function seems to be not created")

        function_name = function["FunctionName"]
        return function_name



import json
import logging
import time

from dataclasses import dataclass, asdict, field
from datetime import datetime
from typing import Optional, List, Tuple

from blazetest.core.project_config.model import BlazetestConfig
from blazetest.core.utils.logging_config import ColoredOutput
from blazetest.core.utils.s3_manager import S3Manager
from blazetest.core.utils.utils import parse_tags

logger = logging.getLogger(__name__)


@dataclass
class JUnitXMLReport:
    test_node_id: str
    report_path: str


@dataclass
class InvocationResult:
    passed_tests_count: int
    failed_tests_count: int

    failed_tests_node_ids: List[str]
    pytest_duration: float

    junit_xml_reports_paths: List[JUnitXMLReport]

    @classmethod
    def parse(
        cls, results: List[Tuple[str, dict]], start_time: float
    ) -> "InvocationResult":
        tests_passed = 0
        failed_tests = []
        junitxml_reports_paths = []

        for node_id, invocation_result in results:
            if invocation_result["test_result"]:
                tests_passed += 1
            else:
                failed_tests.append(node_id)

            junitxml_reports_paths.append(
                JUnitXMLReport(
                    test_node_id=node_id, report_path=invocation_result["report_path"]
                )
            )

        return cls(
            passed_tests_count=tests_passed,
            failed_tests_count=len(failed_tests),
            failed_tests_node_ids=failed_tests,
            pytest_duration=time.time() - start_time,
            junit_xml_reports_paths=junitxml_reports_paths,
        )


@dataclass
class ReportMergeResult:
    final_report_path: str
    flake_report_path: str

    passed: int
    failed: int
    flaky: int

    passed_ids: List[str]
    failed_ids: List[str]
    flaky_ids: List[str]


@dataclass
class TestSessionResult:
    uuid: str

    lambda_function_name: str
    s3_bucket: str

    tests_count: int
    tests_passed: int

    failed_tests_count: int
    flaky_tests_count: int

    pytest_duration: float

    failed_tests_ids: List[str] = field(default_factory=lambda: [])
    flaky_tests_ids: List[str] = field(default_factory=lambda: [])

    start_timestamp: str = None
    end_timestamp: str = None

    tags: str = None

    junit_report_path: str = None
    flake_report_path: str = None

    config: BlazetestConfig = None

    rerun_uuid: str = None
    flake_detected: bool = False

    def set_uuid(self, uuid: str):
        self.uuid = uuid

    def set_rerun_uuid(self, rerun_uuid: str):
        self.rerun_uuid = rerun_uuid

    def set_tags(self, tags: str):
        self.tags = tags

    def log_results(self, failed_test_retry_enabled: bool = False):
        minutes = int(self.pytest_duration // 60)
        seconds = int(self.pytest_duration % 60)
        duration_str = (
            f"{minutes} min {seconds} seconds" if minutes > 0 else f"{seconds} seconds"
        )

        logger.info(f"Tests duration: {duration_str}")
        logger.info(
            f"Failed tests: {self.tests_count - self.tests_passed} out of {self.tests_count}"
        )

        if len(self.failed_tests_ids) > 0:
            print("{:-^50}".format(f"{self.failed_tests_count} FAILED TESTS"))
            for failed_test_id in self.failed_tests_ids:
                print(
                    f"- {ColoredOutput.RED.value}[{failed_test_id}]{ColoredOutput.RESET.value}"
                )

        # If the configuration was set to retry the tests that have failed and license has this feature
        # failed_test_retry is by default 0
        if (
            self.flaky_tests_count > 0
            and failed_test_retry_enabled
            and self.config.general.flaky.remove_flakes is False
        ):
            print("{:-^50}".format(f"{self.flaky_tests_count} FLAKY TESTS"))
            for flaky_test_id in self.flaky_tests_ids:
                print(
                    f"- {ColoredOutput.YELLOW.value}[{flaky_test_id}]{ColoredOutput.RESET.value}"
                )

    def get_tabular_data(self) -> list:
        return [
            self.uuid,
            self.rerun_uuid if self.rerun_uuid else "-",
            self.config["cloud"]["aws"]["resource_prefix"],
            self.__get_normalized_timestamp(),
            f"{self.tests_passed}/{self.tests_count}",
            self.tags if self.tags else "-",
        ]

    def __get_normalized_timestamp(self) -> str:
        return datetime.strptime(self.start_timestamp, "%Y-%m-%d_%H-%M-%S").strftime(
            "%Y-%m-%d %H:%M:%S"
        )


class TestSessionResultManager:
    BLAZETEST_REPORT_FILEPATH = "reports/blazetest-report.json"

    def __init__(
        self,
        s3_manager: S3Manager = None,
        config: BlazetestConfig = None,
    ):
        self.s3_manager = s3_manager
        self.config = config

    def load(self) -> List[TestSessionResult]:
        """
        Load the results from an S3 bucket file.

        Args:
        bucket_name (str): The name of the S3 bucket.
        file_key (str): The key of the file in the S3 bucket.
        """
        results = self.s3_manager.get_json_object(self.BLAZETEST_REPORT_FILEPATH)

        if results is None:
            logger.warning("Could not find the BlazeTest report file in S3 bucket.")
            return []

        if len(results) == 0:
            logger.debug("BlazeTest report is empty")
            return []

        session_results = []
        for result in results:
            session_results.append(TestSessionResult(**result))

        session_results = list(
            sorted(session_results, key=lambda x: x.start_timestamp, reverse=True)
        )
        return session_results

    def append_results_to_json(self, test_session_result: TestSessionResult) -> None:
        """
        Appends the results to a json file in the specified S3 bucket.
        The file is created in the bucket if it does not exist.

        Args:
        bucket_name (str): The name of the S3 bucket.
        file_key (str): The key of the file in the S3 bucket.
        """
        results = self.s3_manager.get_json_object(self.BLAZETEST_REPORT_FILEPATH)

        if results is None:
            # If the report does not exist, we create a new report file
            results = []

        result_dict = asdict(test_session_result)
        results.append(result_dict)

        updated_file_content = json.dumps(results, indent=4)

        uploaded = self.s3_manager.put_object(
            key=self.BLAZETEST_REPORT_FILEPATH,
            body=updated_file_content,
        )

        if uploaded:
            logger.info("Test session results have been appended to file in S3 bucket.")
        else:
            logger.error(
                "Internal error. Could not append test session results to file in S3 bucket."
            )

    def get_latest_test_session_result(self) -> Optional["TestSessionResult"]:
        """
        Gets latest test session result from the json file ("blazetest-report.json") in the project root.
        Returns the dataclass object of the latest test session result.
        """
        results = self.load()
        if not results:
            return None
        results = [result for result in results if result.rerun_uuid is None]
        latest_result = max(results, key=lambda x: x.start_timestamp)
        return latest_result

    def get_test_session_by_uuid(self, uuid: str) -> Optional["TestSessionResult"]:
        """
        Gets test session result from the json file ("blazetest-report.json") in the project root.
        Returns the dataclass object of the test session result.
        """
        results = self.load()

        if not results:
            return None

        results = [
            result
            for result in results
            if result.uuid == uuid and result.rerun_uuid is None
        ]

        if not results:
            return None

        return results[0]

    def get_test_session_results(
        self,
        tags: str,
        include: str = "all",
        uuid: str = None,
    ) -> List["TestSessionResult"]:
        """
        Gets all test session results from the json file ("blazetest-report.json") in the project root.
        Returns the list of dataclass objects of the test session results.
        """
        results = self.load()
        if not results:
            return []

        if include == "runs":
            results = [result for result in results if result.rerun_uuid is None]
        elif include == "reruns":
            results = [result for result in results if result.rerun_uuid is not None]

        if tags:
            results = self.__filter_by_tag(results, tags=tags)

        if uuid:
            results = [result for result in results if result.uuid == uuid]

        return results

    @classmethod
    def __filter_by_tag(
        cls, results: List["TestSessionResult"], tags: str
    ) -> List["TestSessionResult"]:
        """
        Tags are given in the format of tag1=value1,tag2=value2
        If any matches found for the tag, it is added to the list of results.
        """
        requested_tag_dict = parse_tags(tags)

        filtered_results = []
        for result in results:
            if result.tags is None:
                continue
            for key, value in requested_tag_dict.items():
                tag_dict = parse_tags(result.tags)
                if key in tag_dict and tag_dict[key] == value:
                    filtered_results.append(result)

        return filtered_results



import dataclasses
from typing import List, Dict, Optional

from blazetest.core.project_config.validators import (
    ValidationBase,
    ecr_repository_name_is_valid,
    resource_prefix_is_valid,
    s3_bucket_name_is_valid,
    lambda_function_memory_size_is_valid,
    lambda_function_timeout_is_valid,
    junit_results_file_link_is_valid,
)


@dataclasses.dataclass
class LicenceConfig(ValidationBase):
    license_key: Optional[str] = None
    license_file: Optional[str] = None


@dataclasses.dataclass
class FlakeDetectionConfig(ValidationBase):
    failure_retry_attempts: int = 0
    fail_on_flake: bool = True
    remove_flakes: bool = False
    exit_on_flake_detection: bool = True


@dataclasses.dataclass
class GeneralConfig(ValidationBase):
    junit_results_file_link: str = "private"

    license: LicenceConfig = dataclasses.field(default_factory=lambda: LicenceConfig())
    flaky: FlakeDetectionConfig = dataclasses.field(
        default_factory=lambda: FlakeDetectionConfig()
    )

    def get_validators(self) -> Dict:
        return {
            "junit_results_file_link": junit_results_file_link_is_valid,
        }


@dataclasses.dataclass
class AWSConfig(ValidationBase):
    region: str
    resource_prefix: str = "blazetest-stack"
    s3_bucket_prefix: str = "blazetest-s3"
    ecr_repository_prefix: str = "blazetest-repo"

    lambda_function_memory_size: int = 4096
    lambda_function_timeout: int = 900

    tags: dict = dataclasses.field(default_factory=lambda: {})

    def get_resource_name(self, uuid: str):
        return f"{self.resource_prefix}-{uuid}"

    def get_ecr_repository_prefix(self, uuid: str):
        return f"{self.ecr_repository_prefix}-{uuid}"

    def get_validators(self) -> Dict:
        return {
            "ecr_repository_prefix": ecr_repository_name_is_valid,
            "resource_prefix": resource_prefix_is_valid,
            "s3_bucket_prefix": s3_bucket_name_is_valid,
            "lambda_function_memory_size": lambda_function_memory_size_is_valid,
            "lambda_function_timeout": lambda_function_timeout_is_valid,
        }


@dataclasses.dataclass
class AzureConfig(ValidationBase):
    ...


@dataclasses.dataclass
class GCPConfig(ValidationBase):
    ...


@dataclasses.dataclass
class CloudConfig(ValidationBase):
    aws: AWSConfig
    azure: AzureConfig = dataclasses.field(default_factory=lambda: AzureConfig())
    gcp: GCPConfig = dataclasses.field(default_factory=lambda: GCPConfig())


@dataclasses.dataclass
class ChromeConfig(ValidationBase):
    version: str = "latest"


@dataclasses.dataclass
class FirefoxConfig(ValidationBase):
    version: str = "latest"


@dataclasses.dataclass
class EdgeConfig(ValidationBase):
    version: str = "latest"


@dataclasses.dataclass
class BrowserConfig(ValidationBase):
    """
    The list of supported browsers within blazetest docker image can be found at https://blazetest.io/browsers
    """

    firefox: FirefoxConfig = dataclasses.field(default_factory=lambda: FirefoxConfig())
    edge: EdgeConfig = dataclasses.field(default_factory=lambda: EdgeConfig())
    chrome: ChromeConfig = dataclasses.field(default_factory=lambda: ChromeConfig())


@dataclasses.dataclass
class PytestConfig(ValidationBase):
    collection_args: List[str] = dataclasses.field(default_factory=lambda: [])
    execution_args: List[str] = dataclasses.field(default_factory=lambda: [])


@dataclasses.dataclass
class JUnitConfig(ValidationBase):
    ...


@dataclasses.dataclass
class TestingConfig(ValidationBase):
    ...


@dataclasses.dataclass
class FrameworkConfig(ValidationBase):
    pytest: PytestConfig
    junit: JUnitConfig = dataclasses.field(default_factory=lambda: JUnitConfig())
    testing: TestingConfig = dataclasses.field(default_factory=lambda: TestingConfig())


@dataclasses.dataclass
class BlazetestConfig(ValidationBase):
    general: GeneralConfig
    cloud: CloudConfig
    framework: FrameworkConfig
    browser: BrowserConfig = dataclasses.field(default_factory=lambda: BrowserConfig())



import logging
import os
from collections import defaultdict
from typing import List, Dict
from xml.dom.minidom import parseString

import boto3
import junitparser
from junitparser import JUnitXml, TestSuite
from xml.etree import ElementTree

from blazetest.core.config import CWD
from blazetest.core.project_config.model import BlazetestConfig
from blazetest.core.run_test.result_model import JUnitXMLReport, ReportMergeResult
from blazetest.core.utils.logging_config import ColoredOutput
from blazetest.core.utils.exceptions import ReportNotAvailable, ReportNotUploaded

logger = logging.getLogger(__name__)


class ReportMerger:
    """
    Merges reports from S3 Bucket into one file and saves back to the bucket.
    """

    FILE_ENCODING = "utf-8"

    flake_detected: bool = False

    FINAL_REPORT_FILEPATH = (
        "{timestamp}/target/merged/test-session-{resource_prefix}.xml"
    )
    FLAKE_REPORT_FILEPATH = (
        "{timestamp}/target/flake/test-session-{resource_prefix}.xml"
    )

    def __init__(
        self,
        resource_prefix: str,
        region: str,
        s3_bucket_name: str = None,
        config: BlazetestConfig = None,
        aws_access_key_id: str = None,
        aws_secret_access_key: str = None,
    ):
        self.s3_client = boto3.client(
            "s3",
            region_name=region,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )
        self.s3_bucket_name = s3_bucket_name
        self.resource_prefix = resource_prefix
        self.config = config

    def set_s3_bucket_name(self, s3_bucket_name: str) -> None:
        self.s3_bucket_name = s3_bucket_name

    def merge_reports(
        self, reports: List[JUnitXMLReport], timestamp: str
    ) -> ReportMergeResult:
        print(
            f"\n* Downloading {len(reports)} test reports "
            f"{ColoredOutput.GREEN.value}...{ColoredOutput.RESET.value} ",
            end="",
        )
        tests_results = self.get_test_results_by_node_id(reports)
        print(f"{ColoredOutput.GREEN.value}Done{ColoredOutput.RESET.value}")

        print(
            f"* Merging test reports into single JUnitXML test report "
            f"{ColoredOutput.GREEN.value}...{ColoredOutput.RESET.value} ",
            end="",
        )
        merge_result = self.get_final_reports(tests_results)
        print(f"{ColoredOutput.GREEN.value}Done{ColoredOutput.RESET.value}")

        final_report_filepath = self.FINAL_REPORT_FILEPATH.format(
            timestamp=timestamp, resource_prefix=self.resource_prefix
        )
        flake_report_filepath = self.FLAKE_REPORT_FILEPATH.format(
            timestamp=timestamp, resource_prefix=self.resource_prefix
        )

        reports = [
            (merge_result["final_report"], final_report_filepath),
            (merge_result["flake_report"], flake_report_filepath),
        ]

        print(
            f"* Uploading merged JUnitXML test report to S3 bucket "
            f"{ColoredOutput.GREEN.value}...{ColoredOutput.RESET.value} ",
            end="",
        )
        for report, report_path in reports:
            self.__upload_report(
                body=self.formatted_xml_string(report),
                path=report_path,
            )
            with open(os.path.join(CWD, report_path.replace("/", "-")), "w") as f:
                f.write(self.formatted_xml_string(report))

        print(f"{ColoredOutput.GREEN.value}Done{ColoredOutput.RESET.value}\n")

        return ReportMergeResult(
            final_report_path=final_report_filepath,
            flake_report_path=flake_report_filepath,
            passed=merge_result["passed"],
            flaky=merge_result["flaky"],
            failed=merge_result["failed"],
            passed_ids=merge_result["passed_ids"],
            flaky_ids=merge_result["flaky_ids"],
            failed_ids=merge_result["failed_ids"],
        )

    def get_test_results_by_node_id(
        self, reports: List[JUnitXMLReport]
    ) -> Dict[str, dict]:
        tests_results = defaultdict(lambda: defaultdict(list))
        for report in reports:
            report_data = self.__download_report(report.report_path)
            junit_report = junitparser.JUnitXml.fromstring(report_data)

            if junit_report.failures or junit_report.errors:
                tests_results[report.test_node_id]["failed"].append(junit_report)
            else:
                tests_results[report.test_node_id]["passed"].append(junit_report)

        return tests_results

    @staticmethod
    def formatted_xml_string(junit_xml: JUnitXml) -> str:
        xml_str = junit_xml.tostring()
        root = ElementTree.fromstring(xml_str)
        rough_string = ElementTree.tostring(root, encoding="utf-8")
        re_parsed = parseString(rough_string)
        return re_parsed.toprettyxml(indent="  ")

    def get_final_reports(self, tests_results: Dict[str, dict]) -> Dict:
        final_report = junitparser.JUnitXml()
        flake_report = junitparser.JUnitXml()

        passed, flaky, failed = 0, 0, 0
        passed_ids, flaky_ids, failed_ids = [], [], []

        for node_id in tests_results:
            test_result = tests_results[node_id]

            if len(test_result["failed"]) == 0:
                final_report += test_result["passed"][0]
                passed += 1
                passed_ids.append(node_id)
                continue

            elif len(test_result["passed"]) > 0:
                report = self.get_test_suite_with_flake_property(
                    test_result=test_result
                )

                flake_report += report

                if self.config.general.flaky.remove_flakes is False:
                    final_report += report
                    flaky += 1
                    flaky_ids.append(node_id)
                    failed_ids.append(node_id)
                else:
                    failed += 1
                    failed_ids.append(node_id)

                self.flake_detected = True

            else:
                final_report += test_result["failed"][0]
                failed += 1
                failed_ids.append(node_id)

        return {
            "final_report": final_report,
            "flake_report": flake_report,
            "passed": passed,
            "passed_ids": passed_ids,
            "flaky": flaky,
            "flaky_ids": flaky_ids,
            "failed": failed,
            "failed_ids": failed_ids,
        }

    @staticmethod
    def get_test_suite_with_flake_property(test_result: dict):
        test_suites: JUnitXml = test_result["passed"][0]
        for test_suite in test_suites.iterchildren(TestSuite):
            test_suite.add_property("flake", True)
            tests_count = len(test_result["passed"]) + len(test_result["failed"])
            test_suite.add_property(
                "flake_rate", f"{len(test_result['passed'])}/{tests_count}"
            )
        return test_suites

    def __download_report(self, report_path: str) -> str:
        try:
            response = self.s3_client.get_object(
                Bucket=self.s3_bucket_name, Key=report_path
            )
            report_data = response["Body"].read().decode(self.FILE_ENCODING)
            return report_data
        except Exception as e:
            raise ReportNotAvailable(
                f"Error downloading report {report_path}: {str(e)}"
            )

    def __upload_report(self, body: str, path: str) -> None:
        try:
            self.s3_client.put_object(Body=body, Bucket=self.s3_bucket_name, Key=path)
        except Exception as e:
            raise ReportNotUploaded(f"Error uploading report {path} to S3: {str(e)}")
